IMPLEMENTATION MODULE ShdMem;

IMPORT WinDef, WinNT, WinBase;
IMPORT Lib, Str, Lib2;
IMPORT Types, SO, Cfg, Log;
FROM SYSTEM IMPORT BYTE, ADDRESS, ADR;
FROM Storage IMPORT ALLOCATE;

CONST
  mid = Cfg.midShdMem;
  nam = 'Local\ShdMemTst';  (* 'Global\' seems to require admin rights *)
  namMtx = 'Local\ShMemTst_Mtx';

TYPE
  THandle = WinDef.HANDLE;

CONST
  hndInv = WinBase.INVALID_HANDLE_VALUE;  (* 0FFFFFFFFH *)

TYPE
  TPthStr = ARRAY [0..127] OF CHAR;
  TBufDesc = RECORD
    l: CARDINAL;
    s: TPthStr;
    h: THandle;
  END;
  TBuf = POINTER TO TBufDesc;

CONST
  lMaxBuf = 32 * 1024;

VAR
  adrRd, adrWr: ADDRESS;
  mtx: THandle;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, FALSE );
END Message;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE WrHex( VAR b: SO.TBuffer; x-: ARRAY OF BYTE; l: CARDINAL );
CONST
  lMax = 8;
VAR
  i: CARDINAL;
  c: CHAR;
BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > lMax THEN l := lMax END;
  IF l = 0 THEN RETURN END;
  c := b.c.p;
  b.c.p := '0';
  DEC( l );
  FOR i := 0 TO l DO
    SO.WrShtHex( b, Types.TShtCard( x[i] ), 2 );
    IF i < l THEN SO.WrChar( b, ' ' ) END;
  END;
  b.c.p := c;
END WrHex;


PROCEDURE NewMtx( n-: ARRAY OF CHAR ): THandle;
VAR
  h: THandle;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  h := WinBase.CreateMutex( NIL, FALSE, n );
  IF h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewMtx: CreateMutex failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  RETURN h;
END NewMtx;


(* the shared memory is not automatically interlocked by win (tested) *)
PROCEDURE Enter();
VAR
  i: LONGCARD;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  i := WinBase.WaitForSingleObject( mtx, WinBase.INFINITE );
  IF i = MAX( LONGCARD ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  IF i > 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Enter: WaitForSingleObject returned ' );
    SO.WrLngCard( b, i, 1 );
    Fatal( s );
  END;
END Enter;

PROCEDURE Leave();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF ~ WinBase.ReleaseMutex( mtx ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Leave: ReleaseMutex failed, err ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
END Leave;


PROCEDURE Write( x-: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Write: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > len THEN l := len END;
  Enter();
  Lib.Move( ADR( x ), adrWr, l );
  Leave();
  Show();
  RETURN l;
END Write;

PROCEDURE Read( VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..127] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Read: ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, '  ' );
    WrHex( b, x, l );
    Message( s );
  END Show;

BEGIN
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  IF l > len THEN l := len END;
  Enter();
  Lib.Move( adrRd, ADR( x ), l );
  Leave();
  Show();
  RETURN l;
END Read;


PROCEDURE NewBuf( l: LONGCARD; n-: ARRAY OF CHAR ): TBuf;
VAR
  m: TBuf;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  IF l > lMaxBuf THEN Fatal( 'NewBuf: Invalid buffer length' ) END;
  IF Str.Length( n ) >= SIZE( m^.s ) THEN Fatal( 'NewBuf: Name too long' ) END;  (* term rqd *)
  NEW( m );
  Lib2.Fill( m^, 0 );
  m^.l := l;
  Str.Copy( m^.s, n );
  m^.h := WinBase.CreateFileMapping(
    hndInv,                (* file handle *)
    NIL,                   (* attributes *)
    WinNT.PAGE_READWRITE,  (* protection *)
    0,                     (* size high *)
    m^.l,                  (* size low *)
    m^.s                   (* name *)
  );
  IF m^.h = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'NewBuf: CreateFileMapping failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  (* WinError.ERROR_ALREADY_EXISTS = 183 *)
  SO.Init( b, s );
  SO.WrStr( b, 'NewBuf: CreateFileMapping succeeded, error code ' );
  SO.WrLngCard( b, WinBase.GetLastError(), 1 );
  Message( s );
  RETURN m;
END NewBuf;


PROCEDURE Create();
VAR
  m: TBuf;
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  m := NewBuf( len, nam );

  adrRd := WinBase.MapViewOfFile(
    m^.h,
    WinBase.FILE_MAP_READ,
    0,  (* offset high *)
    0,  (* offset low *)
    len (* size (still unclear, since 1 works) *)
  );
  IF adrRd = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Create: MapViewOfFile (read) failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  SO.Init( b, s );
  SO.WrStr( b, 'Create: MapViewOfFile (read) succeeded, adr ' );
  SO.WrLngHex( b, LONGCARD( adrRd ), 8 );
  Message( s );

  adrWr := WinBase.MapViewOfFile(
    m^.h,
    WinBase.FILE_MAP_WRITE,
    0,  (* offset high *)
    0,  (* offset low *)
    len (* size (still unclear, since 1 works) *)
  );
  IF adrWr = NIL THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Create: MapViewOfFile (write) failed, error code ' );
    SO.WrLngCard( b, WinBase.GetLastError(), 1 );
    Fatal( s );
  END;
  SO.Init( b, s );
  SO.WrStr( b, 'Create: MapViewOfFile (write) succeeded, adr ' );
  SO.WrLngHex( b, LONGCARD( adrWr ), 8 );
  Message( s );
END Create;

PROCEDURE InitModule();
BEGIN
  Log.Register( mid, 'ShdMem' );
  Create();
  mtx := NewMtx( namMtx );
END InitModule;

BEGIN
  InitModule();
END ShdMem.
